# Copyright (C) 2019-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""WebGraph driver"""

from datetime import datetime
from enum import Enum
import logging
import os
from pathlib import Path
import subprocess
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Set, Union

# WARNING: do not import unnecessary things here to keep cli startup time under
# control
from swh.graph.config import check_config_compress

from .shell import AtomicFileSink, Command, CommandException, Rust

if TYPE_CHECKING:
    from .shell import RunResult

logger = logging.getLogger(__name__)


class CompressionSubprocessError(Exception):
    def __init__(self, message: str, log_path: Path):
        super().__init__(f"{message}; full logs at {log_path}")
        self.message = message
        self.log_path = log_path


class CompressionStep(Enum):
    EXTRACT_NODES = -20
    EXTRACT_LABELS = -10
    NODE_STATS = 0
    EDGE_STATS = 3
    LABEL_STATS = 6
    MPH = 10
    BV = 30
    BV_EF = 40
    BFS_ROOTS = 50
    BFS = 60
    PERMUTE_AND_SIMPLIFY_BFS = 70
    BFS_EF = 80
    BFS_DCF = 90
    LLP = 100
    COMPOSE_ORDERS = 110
    PERMUTE_LLP = 120
    EF = 140
    TRANSPOSE = 160
    TRANSPOSE_EF = 170
    MAPS = 180
    EXTRACT_PERSONS = 190
    PERSONS_STATS = 195
    MPH_PERSONS = 200
    EXTRACT_FULLNAMES = 203
    FULLNAMES_EF = 207
    NODE_PROPERTIES = 210
    MPH_LABELS = 220
    LABELS_ORDER = 225
    FCL_LABELS = 230
    EDGE_LABELS = 240
    EDGE_LABELS_TRANSPOSE = 250
    EDGE_LABELS_EF = 270
    EDGE_LABELS_TRANSPOSE_EF = 280
    STATS = 290
    E2E_CHECK = 295
    CLEAN_TMP = 300

    def __str__(self):
        return self.name


# full compression pipeline
COMP_SEQ = list(CompressionStep)


def _extract_nodes(conf: Dict[str, Any], env: Dict[str, str]) -> Command:
    return Rust(
        "swh-graph-extract",
        "extract-nodes",
        "--format",
        "orc",
        "--allowed-node-types",
        conf.get("object_types", "*"),
        f"{conf['in_dir']}",
        f"{conf['out_dir']}/{conf['graph_name']}.nodes/",
        conf=conf,
        env=env,
    )


def _extract_labels(conf: Dict[str, Any], env: Dict[str, str]) -> Optional[Command]:
    if {"dir", "snp", "*"}.isdisjoint(set(conf.get("object_types", "*").split(","))):
        return None
    return Rust(
        "swh-graph-extract",
        "extract-labels",
        "--format",
        "orc",
        "--allowed-node-types",
        conf.get("object_types", "*"),
        conf["in_dir"],
        conf=conf,
        env=env,
    ) | Command.zstdmt() > AtomicFileSink(
        Path(f"{conf['out_dir']}/{conf['graph_name']}.labels.csv.zst")
    )


def _node_stats(conf: Dict[str, Any], env: Dict[str, str]) -> Command:
    return Rust(
        "swh-graph-extract",
        "node-stats",
        "--format",
        "orc",
        "--swhids-dir",
        f"{conf['out_dir']}/{conf['graph_name']}.nodes/",
        "--target-stats",
        f"{conf['out_dir']}/{conf['graph_name']}.nodes.stats.txt",
        "--target-count",
        f"{conf['out_dir']}/{conf['graph_name']}.nodes.count.txt",
        conf=conf,
        env=env,
    )


def _edge_stats(conf: Dict[str, Any], env: Dict[str, str]) -> Command:
    return Rust(
        "swh-graph-extract",
        "edge-stats",
        "--format",
        "orc",
        "--allowed-node-types",
        conf.get("object_types", "*"),
        "--dataset-dir",
        f"{conf['in_dir']}",
        "--target-stats",
        f"{conf['out_dir']}/{conf['graph_name']}.edges.stats.txt",
        "--target-count",
        f"{conf['out_dir']}/{conf['graph_name']}.edges.count.txt",
        conf=conf,
        env=env,
    )


def _label_stats(conf: Dict[str, Any], env: Dict[str, str]) -> AtomicFileSink:
    if {"dir", "snp", "*"}.isdisjoint(set(conf.get("object_types", "*").split(","))):
        return Command.echo("0") > AtomicFileSink(
            Path(f"{conf['out_dir']}/{conf['graph_name']}.labels.count.txt")
        )
    return Command.zstdcat(
        f"{conf['out_dir']}/{conf['graph_name']}.labels.csv.zst"
    ) | Command.wc("-l") > AtomicFileSink(
        Path(f"{conf['out_dir']}/{conf['graph_name']}.labels.count.txt")
    )


def _mph(conf: Dict[str, Any], env: Dict[str, str]) -> Command:
    with open(
        f"{conf['out_dir']}/{conf['graph_name']}.nodes.count.txt", "r"
    ) as nodes_count:
        num_nodes = nodes_count.readline().splitlines()
        assert len(num_nodes) == 1
    return Rust(
        "swh-graph-compress",
        "pthash-swhids",
        "--num-nodes",
        num_nodes[0],
        f"{conf['out_dir']}/{conf['graph_name']}.nodes/",
        f"{conf['out_dir']}/{conf['graph_name']}.pthash",
        conf=conf,
        env=env,
    )


def _bv(conf: Dict[str, Any], env: Dict[str, str]) -> Command:
    with open(f"{conf['out_dir']}/{conf['graph_name']}.nodes.count.txt") as nodes_count:
        num_nodes = nodes_count.readline().splitlines()
        assert len(num_nodes) == 1
    batch_size = int(conf.get("batch_size", "0"))
    batching = ["--sort-batch-size", str(batch_size)] if batch_size else []
    return Rust(
        "swh-graph-extract",
        "bv",
        "--allowed-node-types",
        conf.get("object_types", "*"),
        "--mph-algo",
        "pthash",
        "--function",
        f"{conf['out_dir']}/{conf['graph_name']}",
        "--num-nodes",
        num_nodes[0],
        *batching,
        f"{conf['in_dir']}",
        f"{conf['out_dir']}/{conf['graph_name']}-base",
        conf=conf,
        env=env,
    )


def _bv_ef(conf: Dict[str, Any], env: Dict[str, str]) -> Command:
    return Rust(
        "swh-graph-index",
        "ef",
        f"{conf['out_dir']}/{conf['graph_name']}-base",
        conf=conf,
        env=env,
    )


def _bfs_roots(conf: Dict[str, Any], env: Dict[str, str]) -> Command:
    return Rust(
        "swh-graph-extract",
        "bfs-roots",
        "--allowed-node-types",
        conf.get("object_types", "*"),
        f"{conf['in_dir']}",
        f"{conf['out_dir']}/{conf['graph_name']}-bfs.roots.txt",
        conf=conf,
        env=env,
    )


def _bfs(conf: Dict[str, Any], env: Dict[str, str]) -> Command:
    return Rust(
        "swh-graph-compress",
        "bfs",
        "--mph-algo",
        "pthash",
        "--function",
        f"{conf['out_dir']}/{conf['graph_name']}.pthash",
        "--init-roots",
        f"{conf['out_dir']}/{conf['graph_name']}-bfs.roots.txt",
        f"{conf['out_dir']}/{conf['graph_name']}-base",
        f"{conf['out_dir']}/{conf['graph_name']}-bfs.order",
        conf=conf,
        env=env,
    )


def _permute_and_simplify_bfs(conf: Dict[str, Any], env: Dict[str, str]) -> Command:
    batch_size = int(conf.get("batch_size", "0"))
    batching = ["--sort-batch-size", str(batch_size)] if batch_size else []
    return Rust(
        "swh-graph-compress",
        "permute-and-symmetrize",
        f"{conf['out_dir']}/{conf['graph_name']}-base",
        f"{conf['out_dir']}/{conf['graph_name']}-bfs-simplified",
        "--permutation",
        f"{conf['out_dir']}/{conf['graph_name']}-bfs.order",
        *batching,
        conf=conf,
        env=env,
    )


def _bfs_ef(conf: Dict[str, Any], env: Dict[str, str]) -> Command:
    return Rust(
        "swh-graph-index",
        "ef",
        f"{conf['out_dir']}/{conf['graph_name']}-bfs-simplified",
        conf=conf,
        env=env,
    )


def _bfs_dcf(conf: Dict[str, Any], env: Dict[str, str]) -> Command:
    return Rust(
        "swh-graph-index",
        "dcf",
        f"{conf['out_dir']}/{conf['graph_name']}-bfs-simplified",
        conf=conf,
        env=env,
    )


def _llp(conf: Dict[str, Any], env: Dict[str, str]) -> Command:
    return Rust(
        "swh-graph-compress",
        "llp",
        "-g",
        f"{conf['llp_gammas']}",
        f"{conf['out_dir']}/{conf['graph_name']}-bfs-simplified",
        f"{conf['out_dir']}/{conf['graph_name']}-llp.order",
        conf=conf,
        env=env,
    )


def _compose_orders(conf: Dict[str, Any], env: Dict[str, str]) -> Command:
    with open(f"{conf['out_dir']}/{conf['graph_name']}.nodes.count.txt") as nodes_count:
        num_nodes = nodes_count.readline().splitlines()
        assert len(num_nodes) == 1
    return Rust(
        "swh-graph-compress",
        "compose-orders",
        "--num-nodes",
        num_nodes[0],
        "--input",
        f"{conf['out_dir']}/{conf['graph_name']}-bfs.order",
        "--input",
        f"{conf['out_dir']}/{conf['graph_name']}-llp.order",
        "--output",
        f"{conf['out_dir']}/{conf['graph_name']}.pthash.order",
        conf=conf,
        env=env,
    )


def _permute_llp(conf: Dict[str, Any], env: Dict[str, str]) -> Command:
    batch_size = int(conf.get("batch_size", "0"))
    batching = ["--sort-batch-size", str(batch_size)] if batch_size else []
    return Rust(
        "swh-graph-compress",
        "permute",
        f"{conf['out_dir']}/{conf['graph_name']}-base",
        f"{conf['out_dir']}/{conf['graph_name']}",
        "--permutation",
        f"{conf['out_dir']}/{conf['graph_name']}.pthash.order",
        *batching,
        conf=conf,
        env=env,
    )


def _ef(conf: Dict[str, Any], env: Dict[str, str]) -> Command:
    return Rust(
        "swh-graph-index",
        "ef",
        f"{conf['out_dir']}/{conf['graph_name']}",
        conf=conf,
        env=env,
    )


def _transpose(conf: Dict[str, Any], env: Dict[str, str]) -> Command:
    batch_size = int(conf.get("batch_size", "0"))
    batching = ["--sort-batch-size", str(batch_size)] if batch_size else []
    return Rust(
        "swh-graph-compress",
        "transpose",
        f"{conf['out_dir']}/{conf['graph_name']}",
        f"{conf['out_dir']}/{conf['graph_name']}-transposed",
        *batching,
        conf=conf,
        env=env,
    )


def _transpose_ef(conf: Dict[str, Any], env: Dict[str, str]) -> Command:
    return Rust(
        "swh-graph-index",
        "ef",
        f"{conf['out_dir']}/{conf['graph_name']}-transposed",
        conf=conf,
        env=env,
    )


def _maps(conf: Dict[str, Any], env: Dict[str, str]) -> Command:
    with open(f"{conf['out_dir']}/{conf['graph_name']}.nodes.count.txt") as nodes_count:
        num_nodes = nodes_count.readline().splitlines()
        assert len(num_nodes) == 1
    return Rust(
        "swh-graph-compress",
        "maps",
        "--num-nodes",
        num_nodes[0],
        "--swhids-dir",
        f"{conf['out_dir']}/{conf['graph_name']}.nodes/",
        "--mph-algo",
        "pthash",
        "--function",
        f"{conf['out_dir']}/{conf['graph_name']}.pthash",
        "--order",
        f"{conf['out_dir']}/{conf['graph_name']}.pthash.order",
        "--node2swhid",
        f"{conf['out_dir']}/{conf['graph_name']}.node2swhid.bin",
        "--node2type",
        f"{conf['out_dir']}/{conf['graph_name']}.node2type.bin",
        conf=conf,
        env=env,
    )


def _extract_persons(conf: Dict[str, Any], env: Dict[str, str]) -> AtomicFileSink:
    if {"rel", "rev", "*"}.isdisjoint(set(conf.get("object_types", "*").split(","))):
        return Command.echo("") | Command.zstdmt() > AtomicFileSink(
            Path(f"{conf['out_dir']}/{conf['graph_name']}.persons.csv.zst")
        )
    return Rust(
        "swh-graph-extract",
        "extract-persons",
        "--allowed-node-types",
        conf.get("object_types", "*"),
        f"{conf['in_dir']}",
        conf=conf,
        env=env,
    ) | Command.zstdmt() > AtomicFileSink(
        Path(f"{conf['out_dir']}/{conf['graph_name']}.persons.csv.zst")
    )


def _mph_persons(
    conf: Dict[str, Any], env: Dict[str, str]
) -> Union[Command, AtomicFileSink]:
    with open(
        f"{conf['out_dir']}/{conf['graph_name']}.persons.count.txt"
    ) as persons_count:
        num_persons = persons_count.readline().splitlines()
        assert len(num_persons) == 1
    if num_persons[0] == "0":
        return Command.echo("") > AtomicFileSink(
            Path(f"{conf['out_dir']}/{conf['graph_name']}.persons.pthash")
        )
    return Rust(
        "swh-graph-compress",
        "pthash-persons",
        "--num-persons",
        num_persons[0],
        Command.zstdcat(
            f"{conf['out_dir']}/{conf['graph_name']}.persons.csv.zst",
        ),
        f"{conf['out_dir']}/{conf['graph_name']}.persons.pthash",
        conf=conf,
        env=env,
    )


def _extract_fullnames(conf: Dict[str, Any], env: Dict[str, str]) -> Optional[Command]:
    if {"rel", "rev", "*"}.isdisjoint(set(conf.get("object_types", "*").split(","))):
        return None
    if not (
        Path(f"{conf['out_dir']}/{conf['graph_name']}.persons.count.txt").exists()
        and Path(f"{conf['sensitive_in_dir']}/person").exists()
    ):
        return None
    with open(
        f"{conf['out_dir']}/{conf['graph_name']}.persons.count.txt"
    ) as persons_count:
        num_persons = persons_count.readline().splitlines()
        assert len(num_persons) == 1
    if num_persons[0] == "0":
        return None
    return Rust(
        "swh-graph-extract",
        "extract-fullnames",
        "--person-function",
        f"{conf['out_dir']}/{conf['graph_name']}.persons.pthash",
        f"{conf['sensitive_in_dir']}",
        f"{conf['sensitive_out_dir']}/{conf['graph_name']}.persons",
        f"{conf['sensitive_out_dir']}/{conf['graph_name']}.persons.lengths",
    )


def _fullnames_ef(conf: Dict[str, Any], env: Dict[str, str]) -> Optional[Command]:
    if {"rel", "rev", "*"}.isdisjoint(set(conf.get("object_types", "*").split(","))):
        return None
    if not (
        Path(f"{conf['out_dir']}/{conf['graph_name']}.persons.count.txt").exists()
        and Path(f"{conf['sensitive_in_dir']}/person").exists()
    ):
        return None
    with open(
        f"{conf['out_dir']}/{conf['graph_name']}.persons.count.txt"
    ) as persons_count:
        num_persons = persons_count.readline().splitlines()
        assert len(num_persons) == 1
    if num_persons[0] == "0":
        return None
    return Rust(
        "swh-graph-index",
        "fullnames-ef",
        "--num-persons",
        num_persons[0],
        f"{conf['sensitive_out_dir']}/{conf['graph_name']}.persons",
        f"{conf['sensitive_out_dir']}/{conf['graph_name']}.persons.lengths",
        f"{conf['sensitive_out_dir']}/{conf['graph_name']}.persons.ef",
    )


def _persons_stats(conf: Dict[str, Any], env: Dict[str, str]) -> AtomicFileSink:
    if {"rel", "rev", "*"}.isdisjoint(set(conf.get("object_types", "*").split(","))):
        return Command.echo("0") > AtomicFileSink(
            Path(f"{conf['out_dir']}/{conf['graph_name']}.persons.count.txt")
        )
    return Command.zstdcat(
        f"{conf['out_dir']}/{conf['graph_name']}.persons.csv.zst"
    ) | Command.wc("-l") > AtomicFileSink(
        Path(f"{conf['out_dir']}/{conf['graph_name']}.persons.count.txt")
    )


def _node_properties(conf: Dict[str, Any], env: Dict[str, str]) -> Command:
    with open(f"{conf['out_dir']}/{conf['graph_name']}.nodes.count.txt") as nodes_count:
        num_nodes = nodes_count.readline().splitlines()
        assert len(num_nodes) == 1
    return Rust(
        "swh-graph-extract",
        "node-properties",
        "--format",
        "orc",
        "--allowed-node-types",
        conf.get("object_types", "*"),
        "--mph-algo",
        "pthash",
        "--function",
        f"{conf['out_dir']}/{conf['graph_name']}",
        "--order",
        f"{conf['out_dir']}/{conf['graph_name']}.pthash.order",
        "--person-function",
        f"{conf['out_dir']}/{conf['graph_name']}.persons.pthash",
        "--num-nodes",
        num_nodes[0],
        f"{conf['in_dir']}",
        f"{conf['out_dir']}/{conf['graph_name']}",
        conf=conf,
        env=env,
    )


def _mph_labels(conf: Dict[str, Any], env: Dict[str, str]) -> Optional[Command]:
    if {"dir", "snp", "*"}.isdisjoint(set(conf.get("object_types", "*").split(","))):
        return None
    with open(
        f"{conf['out_dir']}/{conf['graph_name']}.labels.count.txt"
    ) as labels_count:
        num_labels = labels_count.readline().splitlines()
        assert len(num_labels) == 1
    if num_labels[0] == "0":
        return None
    return Rust(
        "swh-graph-compress",
        "pthash-labels",
        "--num-labels",
        num_labels[0],
        Command.zstdcat(f"{conf['out_dir']}/{conf['graph_name']}.labels.csv.zst"),
        f"{conf['out_dir']}/{conf['graph_name']}.labels.pthash",
        conf=conf,
        env=env,
    )


def _labels_order(conf: Dict[str, Any], env: Dict[str, str]) -> Optional[Command]:
    if {"dir", "snp", "*"}.isdisjoint(set(conf.get("object_types", "*").split(","))):
        return None
    with open(
        f"{conf['out_dir']}/{conf['graph_name']}.labels.count.txt"
    ) as labels_count:
        num_labels = labels_count.readline().splitlines()
        assert len(num_labels) == 1
    if num_labels[0] == "0":
        return None
    return Rust(
        "swh-graph-compress",
        "pthash-labels-order",
        "--num-labels",
        num_labels[0],
        Command.zstdcat(f"{conf['out_dir']}/{conf['graph_name']}.labels.csv.zst"),
        f"{conf['out_dir']}/{conf['graph_name']}.labels.pthash",
        f"{conf['out_dir']}/{conf['graph_name']}.labels.pthash.order",
        conf=conf,
        env=env,
    )


def _fcl_labels(conf: Dict[str, Any], env: Dict[str, str]) -> Optional[Command]:
    if {"dir", "snp", "*"}.isdisjoint(set(conf.get("object_types", "*").split(","))):
        return None
    with open(
        f"{conf['out_dir']}/{conf['graph_name']}.labels.count.txt"
    ) as labels_count:
        num_labels = labels_count.readline().splitlines()
        assert len(num_labels) == 1
    return Rust(
        "swh-graph-compress",
        "fcl",
        "--num-lines",
        num_labels[0],
        Command.zstdcat(f"{conf['out_dir']}/{conf['graph_name']}.labels.csv.zst"),
        f"{conf['out_dir']}/{conf['graph_name']}.labels.fcl",
        conf=conf,
        env=env,
    )


def _edge_labels(conf: Dict[str, Any], env: Dict[str, str]) -> Optional[Command]:
    if {"dir", "snp", "ori", "*"}.isdisjoint(
        set(conf.get("object_types", "*").split(","))
    ):
        return None
    with open(
        f"{conf['out_dir']}/{conf['graph_name']}.labels.count.txt"
    ) as labels_count:
        num_labels = labels_count.readline().splitlines()
        assert len(num_labels) == 1
    if num_labels[0] == "0":
        return None
    with open(f"{conf['out_dir']}/{conf['graph_name']}.nodes.count.txt") as nodes_count:
        num_nodes = nodes_count.readline().splitlines()
        assert len(num_nodes) == 1
    if num_nodes[0] == "0":
        return None
    batch_size = int(conf.get("batch_size", "0"))
    batching = ["--sort-batch-size", str(batch_size)] if batch_size else []
    return Rust(
        "swh-graph-extract",
        "edge-labels",
        "--allowed-node-types",
        conf.get("object_types", "*"),
        "--mph-algo",
        "pthash",
        "--function",
        f"{conf['out_dir']}/{conf['graph_name']}",
        "--order",
        f"{conf['out_dir']}/{conf['graph_name']}.pthash.order",
        "--label-name-mphf",
        f"{conf['out_dir']}/{conf['graph_name']}.labels.pthash",
        "--label-name-order",
        f"{conf['out_dir']}/{conf['graph_name']}.labels.pthash.order",
        "--num-nodes",
        num_nodes[0],
        *batching,
        f"{conf['in_dir']}",
        f"{conf['out_dir']}/{conf['graph_name']}",
        conf=conf,
        env=env,
    )


def _edge_labels_transpose(
    conf: Dict[str, Any], env: Dict[str, str]
) -> Optional[Command]:
    if {"dir", "snp", "ori", "*"}.isdisjoint(
        set(conf.get("object_types", "*").split(","))
    ):
        return None
    with open(
        f"{conf['out_dir']}/{conf['graph_name']}.labels.count.txt"
    ) as labels_count:
        num_labels = labels_count.readline().splitlines()
        assert len(num_labels) == 1
    if num_labels[0] == "0":
        return None
    with open(f"{conf['out_dir']}/{conf['graph_name']}.nodes.count.txt") as nodes_count:
        num_nodes = nodes_count.readline().splitlines()
        assert len(num_nodes) == 1
    if num_nodes[0] == "0":
        return None
    batch_size = int(conf.get("batch_size", "0"))
    batching = ["--sort-batch-size", str(batch_size)] if batch_size else []
    return Rust(
        "swh-graph-extract",
        "edge-labels",
        "--allowed-node-types",
        conf.get("object_types", "*"),
        "--mph-algo",
        "pthash",
        "--function",
        f"{conf['out_dir']}/{conf['graph_name']}",
        "--order",
        f"{conf['out_dir']}/{conf['graph_name']}.pthash.order",
        "--label-name-mphf",
        f"{conf['out_dir']}/{conf['graph_name']}.labels.pthash",
        "--label-name-order",
        f"{conf['out_dir']}/{conf['graph_name']}.labels.pthash.order",
        "--num-nodes",
        num_nodes[0],
        "--transposed",
        *batching,
        f"{conf['in_dir']}",
        f"{conf['out_dir']}/{conf['graph_name']}-transposed",
        conf=conf,
        env=env,
    )


def _edge_labels_ef(conf: Dict[str, Any], env: Dict[str, str]) -> Optional[Command]:
    if {"dir", "snp", "ori", "*"}.isdisjoint(
        set(conf.get("object_types", "*").split(","))
    ):
        return None
    with open(
        f"{conf['out_dir']}/{conf['graph_name']}.labels.count.txt"
    ) as labels_count:
        num_labels = labels_count.readline().splitlines()
        assert len(num_labels) == 1
    if num_labels[0] == "0":
        return None
    with open(f"{conf['out_dir']}/{conf['graph_name']}.nodes.count.txt") as nodes_count:
        num_nodes = nodes_count.readline().splitlines()
        assert len(num_nodes) == 1
    return Rust(
        "swh-graph-index",
        "labels-ef",
        f"{conf['out_dir']}/{conf['graph_name']}-labelled",
        num_nodes[0],
        conf=conf,
        env=env,
    )


def _edge_labels_transpose_ef(
    conf: Dict[str, Any], env: Dict[str, str]
) -> Optional[Command]:
    if {"dir", "snp", "*"}.isdisjoint(set(conf.get("object_types", "*").split(","))):
        return None
    with open(
        f"{conf['out_dir']}/{conf['graph_name']}.labels.count.txt"
    ) as labels_count:
        num_labels = labels_count.readline().splitlines()
        assert len(num_labels) == 1
    if num_labels[0] == "0":
        return None
    with open(f"{conf['out_dir']}/{conf['graph_name']}.nodes.count.txt") as nodes_count:
        num_nodes = nodes_count.readline().splitlines()
        assert len(num_nodes) == 1
    return Rust(
        "swh-graph-index",
        "labels-ef",
        f"{conf['out_dir']}/{conf['graph_name']}-transposed-labelled",
        num_nodes[0],
        conf=conf,
        env=env,
    )


def _stats(conf: Dict[str, Any], env: Dict[str, str]) -> Command:
    return Rust(
        "swh-graph-compress",
        "stats",
        "--graph",
        f"{conf['out_dir']}/{conf['graph_name']}",
        "--stats",
        f"{conf['out_dir']}/{conf['graph_name']}.stats",
        conf=conf,
        env=env,
    )


def _e2e_check(
    conf: Dict[str, Any], env: Dict[str, str]
) -> Callable[[logging.Logger], None]:
    from swh.graph.e2e_check import run_e2e_check

    return lambda logger: run_e2e_check(
        graph_name=conf["graph_name"],
        in_dir=conf["in_dir"],
        out_dir=conf["out_dir"],
        sensitive_in_dir=conf.get("sensitive_in_dir"),
        sensitive_out_dir=conf.get("sensitive_out_dir"),
        check_flavor=conf.get("check_flavor", "full"),
        profile=conf.get("profile", "release"),
        logger=logger,
    )


def _clean_tmp(conf: Dict[str, Any], env: Dict[str, str]) -> Command:
    return Command.rm(
        "-rf",
        f"{conf['out_dir']}/{conf['graph_name']}-base.*",
        f"{conf['out_dir']}/{conf['graph_name']}-bfs-simplified.*",
        f"{conf['out_dir']}/{conf['graph_name']}-bfs.order",
        f"{conf['out_dir']}/{conf['graph_name']}-llp.order",
        f"{conf['out_dir']}/{conf['graph_name']}.nodes/",
        f"{conf['out_dir']}/{conf['graph_name']}-bfs.roots.txt",
        f"{conf['out_dir']}/{conf['graph_name']}.labels.csv.zst",
        f"{conf['out_dir']}/{conf['graph_name']}.persons.csv.zst",
        f"{conf['tmp_dir']}",
    )


COMP_CMD: Dict[
    CompressionStep,
    Union[
        Callable[[dict, dict], Optional[Command]],
        Callable[[dict, dict], Optional[AtomicFileSink]],
        Callable[[dict, dict], Union[Command, AtomicFileSink]],
        Callable[[dict, dict], Callable[[logging.Logger], None]],
    ],
] = {
    CompressionStep.EXTRACT_NODES: _extract_nodes,
    CompressionStep.EXTRACT_LABELS: _extract_labels,
    CompressionStep.NODE_STATS: _node_stats,
    CompressionStep.EDGE_STATS: _edge_stats,
    CompressionStep.LABEL_STATS: _label_stats,
    CompressionStep.MPH: _mph,
    CompressionStep.BV: _bv,
    CompressionStep.BV_EF: _bv_ef,
    CompressionStep.BFS_ROOTS: _bfs_roots,
    CompressionStep.BFS: _bfs,
    CompressionStep.PERMUTE_AND_SIMPLIFY_BFS: _permute_and_simplify_bfs,
    CompressionStep.BFS_EF: _bfs_ef,
    CompressionStep.BFS_DCF: _bfs_dcf,
    CompressionStep.LLP: _llp,
    CompressionStep.COMPOSE_ORDERS: _compose_orders,
    CompressionStep.PERMUTE_LLP: _permute_llp,
    CompressionStep.EF: _ef,
    CompressionStep.TRANSPOSE: _transpose,
    CompressionStep.TRANSPOSE_EF: _transpose_ef,
    CompressionStep.MAPS: _maps,
    CompressionStep.EXTRACT_PERSONS: _extract_persons,
    CompressionStep.MPH_PERSONS: _mph_persons,
    CompressionStep.EXTRACT_FULLNAMES: _extract_fullnames,
    CompressionStep.FULLNAMES_EF: _fullnames_ef,
    CompressionStep.PERSONS_STATS: _persons_stats,
    CompressionStep.NODE_PROPERTIES: _node_properties,
    CompressionStep.MPH_LABELS: _mph_labels,
    CompressionStep.LABELS_ORDER: _labels_order,
    CompressionStep.FCL_LABELS: _fcl_labels,
    CompressionStep.EDGE_LABELS: _edge_labels,
    CompressionStep.EDGE_LABELS_TRANSPOSE: _edge_labels_transpose,
    CompressionStep.EDGE_LABELS_EF: _edge_labels_ef,
    CompressionStep.EDGE_LABELS_TRANSPOSE_EF: _edge_labels_transpose_ef,
    CompressionStep.STATS: _stats,
    CompressionStep.E2E_CHECK: _e2e_check,
    CompressionStep.CLEAN_TMP: _clean_tmp,
}


def do_step(step, conf, env=None) -> "List[RunResult]":
    if env is None:
        env = os.environ.copy()
        env["TMPDIR"] = conf["tmp_dir"]
        env["TZ"] = "UTC"

    log_dir = Path(conf["out_dir"]) / "logs"
    log_dir.mkdir(exist_ok=True)

    step_logger = logger.getChild(f"steps.{step.name.lower()}")
    if not step_logger.isEnabledFor(logging.INFO):
        # Ensure that at least INFO messages are sent, because it is the level we use
        # for the stdout of Rust processes. These processes can take a long time to
        # run, and it would be very annoying to have to run them again just because
        # they crashed with no log.
        step_logger.setLevel(logging.INFO)
    log_path = log_dir / (
        f"{conf['graph_name']}"
        f"-{int(datetime.now().timestamp() * 1000)}"
        f"-{str(step).lower()}.log"
    )
    step_handler = logging.FileHandler(log_path)
    step_logger.addHandler(step_handler)

    step_start_time = datetime.now()
    step_logger.info("Starting compression step %s at %s", step, step_start_time)

    command = COMP_CMD[step](conf, env)

    if command is None:
        step_logger.info("Compression step %s skipped", step)
        step_logger.removeHandler(step_handler)
        step_handler.close()
        return []

    step_start_time = datetime.now()
    step_logger.info("Starting compression step %s at %s", step, step_start_time)
    step_logger.info("Running: %s", command.__str__())

    if isinstance(command, (Command, AtomicFileSink)):
        running_command = command._run(
            stdin=None, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )
        with running_command.stdout() as stdout:
            for line in stdout:
                step_logger.info(line.rstrip().decode(errors="replace"))
        try:
            results = running_command.wait()
        except CommandException as e:
            msg = f"Compression step {step} returned non-zero exit code {e.returncode}"
            step_logger.critical(msg)
            raise CompressionSubprocessError(msg, log_path)
        step_end_time = datetime.now()
        step_duration = step_end_time - step_start_time
        step_logger.info(
            "Compression step %s finished at %s (in %s)",
            step,
            step_end_time,
            step_duration,
        )
        step_logger.removeHandler(step_handler)
        step_handler.close()
    else:
        # This allows for calling Python functions directly
        try:
            results = command(step_logger)
        except Exception as exc:
            msg = f"Compression step {step} failed with the following error: {exc}"
            step_logger.critical(msg)
            raise CompressionSubprocessError(msg, log_path)

    return results


def compress(
    graph_name: str,
    in_dir: str,
    out_dir: str,
    sensitive_in_dir: Optional[str],
    sensitive_out_dir: Optional[str],
    check_flavor: Optional[str],
    steps: Set[CompressionStep] = set(COMP_SEQ),
    conf: Dict[str, str] = {},
    progress_cb: Callable[[int, CompressionStep], None] = lambda percentage, step: None,
):
    """graph compression pipeline driver from nodes/edges files to compressed
    on-disk representation

    Args:
        graph_name: graph base name, relative to in_dir
        in_dir: input directory, where the uncompressed graph can be found
        out_dir: output directory, where the compressed graph will be stored
        sensitive_in_dir: sensitive input directory, where the uncompressed
            sensitive graph can be found
        sensitive_out_dir: sensitive output directory, where the compressed
            sensitive graph will be stored
        check_flavor: which flavor of checks to run
        steps: compression steps to run (default: all steps)
        conf: compression configuration, supporting the following keys (all are
          optional, so an empty configuration is fine and is the default)

          - batch_size: batch size for `WebGraph transformations
            <http://webgraph.di.unimi.it/docs/it/unimi/dsi/webgraph/Transform.html>`_;
            defaults to 1 billion
          - tmp_dir: temporary directory, defaults to the "tmp" subdir of
            out_dir
          - object_types: comma-separated list of object types to extract
            (eg. ``ori,snp,rel,rev``). Defaults to ``*``.
        progress_cb: a callable taking a percentage and step as argument,
          which is called every time a step starts.

    """
    if not steps:
        steps = set(COMP_SEQ)

    conf = check_config_compress(
        conf,
        graph_name,
        in_dir,
        out_dir,
        sensitive_in_dir,
        sensitive_out_dir,
        check_flavor,
    )

    compression_start_time = datetime.now()
    logger.info("Starting compression at %s", compression_start_time)
    seq_no = 0
    for step in COMP_SEQ:
        if step not in steps:
            logger.debug("Skipping compression step %s", step)
            continue
        seq_no += 1
        logger.info("Running compression step %s (%s/%s)", step, seq_no, len(steps))
        progress_cb(seq_no * 100 // len(steps), step)
        do_step(step, conf)
    compression_end_time = datetime.now()
    compression_duration = compression_end_time - compression_start_time
    logger.info("Completed compression in %s", compression_duration)



# Copyright (c) 2022-2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import copy
import hashlib

import grpc
import pytest

from swh.graph import pytest_plugin
from swh.graph.grpc.swhgraph_pb2 import GetNodeRequest, NodeFilter, TraversalRequest

TEST_ORIGIN_ID = "swh:1:ori:{}".format(
    hashlib.sha1(b"https://example.com/swh/graph").hexdigest()
)
TEST_ORIGIN_ID2 = "swh:1:ori:{}".format(
    hashlib.sha1(b"https://example.com/swh/graph2").hexdigest()
)


@pytest.fixture(scope="session")
def graph_grpc_server_config(graph_grpc_server_config, tmp_path_factory):
    masked_swhids_path = tmp_path_factory.mktemp("test_masking") / "masked_nodes.txt"
    masked_swhids_path.write_text(
        "\n".join(
            [
                TEST_ORIGIN_ID,
                "swh:1:cnt:0000000000000000000000000000000000000005",
                "swh:1:rel:0000000000000000000000000000000000000010",
                "swh:1:snp:ffffffffffffffffffffffffffffffffffffffff",  # not in the graph
            ]
        )
    )
    config = copy.deepcopy(graph_grpc_server_config)
    config["graph"]["grpc_server"]["masked_nodes"] = str(masked_swhids_path)
    yield config


@pytest.fixture(scope="module")
def graph_grpc_server_process(graph_grpc_server_config, graph_statsd_server):
    # override session-scoped fixture to force a new graph process with the new config
    yield from pytest_plugin.graph_grpc_server_process.__wrapped__(
        graph_grpc_server_config, graph_statsd_server
    )


@pytest.fixture(scope="module")
def graph_grpc_server_started(graph_grpc_server_process):
    # override session-scoped fixture to force a new graph process with the new config
    yield from pytest_plugin.graph_grpc_server_started.__wrapped__(
        graph_grpc_server_process
    )


def test_not_found(graph_grpc_stub, graph_grpc_backend_implementation):
    with pytest.raises(grpc.RpcError) as excinfo:
        graph_grpc_stub.GetNode(
            GetNodeRequest(swhid="swh:1:cnt:0000000000000000000000000000000000000194")
        )
    assert excinfo.value.code() == grpc.StatusCode.NOT_FOUND

    with pytest.raises(grpc.RpcError) as excinfo:
        graph_grpc_stub.GetNode(
            GetNodeRequest(swhid="swh:1:snp:ffffffffffffffffffffffffffffffffffffffff")
        )
    assert excinfo.value.code() == grpc.StatusCode.NOT_FOUND


def test_invalid_swhid(graph_grpc_stub, graph_grpc_backend_implementation):
    with pytest.raises(grpc.RpcError) as excinfo:
        graph_grpc_stub.GetNode(
            GetNodeRequest(swhid="swh:1:lol:0000000000000000000000000000000000000001")
        )
    assert excinfo.value.code() == grpc.StatusCode.INVALID_ARGUMENT

    with pytest.raises(grpc.RpcError) as excinfo:
        graph_grpc_stub.GetNode(
            GetNodeRequest(swhid="swh:1:cnt:000000000000000000000000000000000000000z")
        )
    assert excinfo.value.code() == grpc.StatusCode.INVALID_ARGUMENT


def test_masked_swhid(graph_grpc_stub, graph_grpc_backend_implementation):
    with pytest.raises(grpc.RpcError) as excinfo:
        graph_grpc_stub.GetNode(GetNodeRequest(swhid=TEST_ORIGIN_ID))
    assert excinfo.value.code() == grpc.StatusCode.NOT_FOUND
    assert "Unavailable node" in excinfo.value.details()

    with pytest.raises(grpc.RpcError) as excinfo:
        graph_grpc_stub.GetNode(
            GetNodeRequest(swhid="swh:1:cnt:0000000000000000000000000000000000000005")
        )
    assert excinfo.value.code() == grpc.StatusCode.NOT_FOUND


def test_getnode(graph_grpc_stub, graph_grpc_backend_implementation):
    expected_cnts = [1, 4, 7, 11, 14, 15]

    for cnt_id in expected_cnts:
        graph_grpc_stub.GetNode(GetNodeRequest(swhid=f"swh:1:cnt:{cnt_id:040}"))


def test_traverse(graph_grpc_stub, graph_grpc_backend_implementation):
    request = graph_grpc_stub.Traverse(
        TraversalRequest(
            src=["swh:1:dir:0000000000000000000000000000000000000008"],
            return_nodes=NodeFilter(max_traversal_successors=0),
        )
    )
    assert {node.swhid for node in request} == {
        "swh:1:cnt:0000000000000000000000000000000000000001",
        "swh:1:cnt:0000000000000000000000000000000000000004",
        # "swh:1:cnt:0000000000000000000000000000000000000005",  # masked
        "swh:1:cnt:0000000000000000000000000000000000000007",
    }


def test_traverse_from_masked(graph_grpc_stub, graph_grpc_backend_implementation):
    with pytest.raises(grpc.RpcError) as excinfo:
        list(
            graph_grpc_stub.Traverse(
                TraversalRequest(
                    src=["swh:1:rel:0000000000000000000000000000000000000010"]
                )
            )
        )
    assert excinfo.value.code() == grpc.StatusCode.NOT_FOUND
    assert "Unavailable node" in excinfo.value.details()



# Copyright (C) 2019-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
from pathlib import Path
import shutil
from tempfile import TemporaryDirectory
from typing import Dict

from click.testing import CliRunner
import pytest
import yaml

from swh.graph.cli import graph_cli_group
from swh.graph.example_dataset import DATASET_DIR


def read_properties(properties_fname) -> Dict[str, str]:
    """read a Java .properties file"""
    with open(properties_fname) as f:
        keyvalues = (
            line.split("=", maxsplit=1)
            for line in f
            if not line.strip().startswith("#")
        )
        return dict((k.strip(), v.strip()) for (k, v) in keyvalues)


def test_pipeline():
    """run full compression pipeline"""
    # bare bone configuration, to allow testing the compression pipeline
    # with minimum RAM requirements on trivial graphs
    config = {
        "profile": "debug",
        "graph": {
            "compress": {
                "batch_size": 1000,
                "rust_executable_dir": "./target/debug/",
                "check_flavor": "example",
            }
        },
    }
    runner = CliRunner()

    with TemporaryDirectory(suffix=".swh-graph-test") as tmpdir:
        config_path = Path(tmpdir, "config.yml")
        config_path.write_text(yaml.dump(config))

        result = runner.invoke(
            graph_cli_group,
            [
                "--config-file",
                config_path,
                "compress",
                "--input-dataset",
                DATASET_DIR / "orc",
                "--output-directory",
                tmpdir,
                "--sensitive-output-directory",
                tmpdir,
                "--graph-name",
                "example",
            ],
        )
        assert result.exit_code == 0, result
        properties = read_properties(Path(tmpdir) / "example.properties")

    assert int(properties["nodes"]) == 24
    assert int(properties["arcs"]) == 28


@pytest.mark.parametrize("option", ["none", "--ef", "--force"])
def test_reindex(mocker, tmpdir, option):
    config = {
        "graph": {
            "compress": {
                "rust_executable_dir": "./target/debug/",
            }
        }
    }
    runner = CliRunner()

    trailing = []
    if option != "none":
        trailing.append(option)

    with TemporaryDirectory(suffix=".swh-graph-test") as tmpdir:
        config_path = Path(tmpdir, "config.yml")
        config_path.write_text(yaml.dump(config))

        shutil.copytree(DATASET_DIR / "compressed", tmpdir, dirs_exist_ok=True)

        result = runner.invoke(
            graph_cli_group,
            [
                "--config-file",
                config_path,
                "--profile",
                "debug",
                "reindex",
                f"{tmpdir}/example",
                *trailing,
            ],
        )
        assert result.exit_code == 0, result


@pytest.mark.parametrize("exit_code", [0, 1])
def test_luigi(mocker, tmpdir, exit_code):
    """calls Luigi with the given configuration"""
    # bare bone configuration, to allow testing the compression pipeline
    # with minimum RAM requirements on trivial graphs
    runner = CliRunner()

    subprocess_run = mocker.patch("subprocess.run")
    subprocess_run.return_value.returncode = exit_code

    with TemporaryDirectory(suffix=".swh-graph-test") as tmpdir:
        result = runner.invoke(
            graph_cli_group,
            [
                "luigi",
                "--base-directory",
                f"{tmpdir}/base_dir",
                "--dataset-name",
                "2022-12-07",
                "--",
                "foo",
                "bar",
                "--baz",
                "qux",
            ],
            catch_exceptions=False,
        )
        assert result.exit_code == exit_code, result

    luigi_config_path = subprocess_run.mock_calls[0][2]["env"]["LUIGI_CONFIG_PATH"]
    subprocess_run.assert_called_once_with(
        [
            "luigi",
            "--module",
            "swh.export.luigi",
            "--module",
            "swh.graph.luigi",
            "foo",
            "bar",
            "--baz",
            "qux",
        ],
        env={"LUIGI_CONFIG_PATH": luigi_config_path, **os.environ},
    )



# Copyright (c) 2022-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Iterable, List

import pytest

from swh.graph.example_dataset import DATASET, DATASET_DIR
from swh.graph.shell import Command, CommandException, Rust, Sink


def hash_swhids_with_node2swhid(swhids: Iterable[str]) -> List[int]:
    # fmt: off
    return (
        Command.echo("\n".join(swhids))
        | Rust(
            "swh-graph-hash",
            "swhids",
            "--num-nodes",
            (DATASET_DIR / "compressed/example.nodes.count.txt").read_text().strip(),
            "--mph-algo",
            "pthash",
            "--mph",
            DATASET_DIR / "compressed/example",
            "--permutation",
            DATASET_DIR / "compressed/example.pthash.order",
            "--node2swhid",
            DATASET_DIR / "compressed/example.node2swhid.bin"
        )
        > Sink()
    ).run().stdout.decode().strip().split("\n")
    # fmt: on


def test_hash_all_swhids() -> None:
    swhids = [obj.swhid() for obj in DATASET if hasattr(obj, "swhid")]
    hashes = hash_swhids_with_node2swhid(map(str, swhids))

    # Check there are as many output lines as input lines, and they are each a unique
    # integer
    assert sorted(map(int, hashes)) == list(range(len(swhids)))


def test_hash_unknown_swhid() -> None:
    swhids = ["swh:1:dir:9999999999999999999999999999999999999999"]

    with pytest.raises(CommandException):
        hash_swhids_with_node2swhid(map(str, swhids))



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import signal

import pytest

from swh.core.api import TransientRemoteException
from swh.graph.http_client import RemoteGraphClient
from swh.graph.pytest_plugin import GraphServerProcess

from .test_http_client import TEST_ORIGIN_ID


def test_leaves(graph_grpc_server_config):
    server = GraphServerProcess(graph_grpc_server_config)
    server.start()
    try:
        graph_client = RemoteGraphClient(str(server.result["server_url"]))

        list(graph_client.leaves(TEST_ORIGIN_ID))

        pid = server.result["pid"]
        os.kill(pid, signal.SIGKILL)
        try:
            os.waitpid(pid, os.WNOHANG)
        except ChildProcessError:
            pass

        it = graph_client.leaves(TEST_ORIGIN_ID)
        with pytest.raises(TransientRemoteException, match="failed to connect"):
            list(it)
    finally:
        server.stop()



# Copyright (C) 2019-2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from click.testing import CliRunner

from swh.graph.cli import graph_cli_group


def test_find_context_content(graph_grpc_server):
    swhid = "swh:1:cnt:0000000000000000000000000000000000000001"
    runner = CliRunner()
    result = runner.invoke(
        graph_cli_group, ["find-context", "-g", graph_grpc_server, "-c", swhid]
    )
    expected_fqswhid = (
        "swh:1:cnt:0000000000000000000000000000000000000001;"
        "path=/oldproject/README.md;"
        "anchor=swh:1:rev:0000000000000000000000000000000000000009;"
        "visit=swh:1:snp:0000000000000000000000000000000000000020;"
        "origin=https://example.com/swh/graph\n"
    )
    assert result.exit_code == 0, result
    assert result.output == expected_fqswhid, result.output


def test_find_context_directory(graph_grpc_server):
    swhid = "swh:1:dir:0000000000000000000000000000000000000012"
    runner = CliRunner()
    result = runner.invoke(
        graph_cli_group, ["find-context", "-g", graph_grpc_server, "-c", swhid]
    )
    expected_fqswhid = (
        "swh:1:dir:0000000000000000000000000000000000000012;path=/;"
        "anchor=swh:1:rev:0000000000000000000000000000000000000013;"
        "visit=swh:1:snp:0000000000000000000000000000000000000022;"
        "origin=https://example.com/swh/graph2\n"
    )
    assert result.exit_code == 0, result
    assert result.output == expected_fqswhid, result.output


def test_find_context_revision(graph_grpc_server):
    swhid = "swh:1:rev:0000000000000000000000000000000000000009"
    runner = CliRunner()
    result = runner.invoke(
        graph_cli_group, ["find-context", "-g", graph_grpc_server, "-c", swhid]
    )
    expected_fqswhid = (
        "swh:1:rev:0000000000000000000000000000000000000009;"
        "visit=swh:1:snp:0000000000000000000000000000000000000020;"
        "origin=https://example.com/swh/graph\n"
    )
    assert result.exit_code == 0, result
    assert result.output == expected_fqswhid, result.output


def test_find_context_release(graph_grpc_server):
    swhid = "swh:1:rel:0000000000000000000000000000000000000010"
    runner = CliRunner()
    result = runner.invoke(
        graph_cli_group, ["find-context", "-g", graph_grpc_server, "-c", swhid]
    )
    expected_fqswhid = (
        "swh:1:rel:0000000000000000000000000000000000000010;"
        "visit=swh:1:snp:0000000000000000000000000000000000000020;"
        "origin=https://example.com/swh/graph\n"
    )
    assert result.exit_code == 0, result
    assert result.output == expected_fqswhid, result.output


def test_find_context_snapshot(graph_grpc_server):
    swhid = "swh:1:snp:0000000000000000000000000000000000000020"
    runner = CliRunner()
    result = runner.invoke(
        graph_cli_group, ["find-context", "-g", graph_grpc_server, "-c", swhid]
    )
    expected_fqswhid = (
        "swh:1:snp:0000000000000000000000000000000000000020;"
        "origin=https://example.com/swh/graph\n"
    )
    assert result.exit_code == 0, result
    assert result.output == expected_fqswhid, result.output


def test_find_context_unknown(graph_grpc_server):
    swhid = "swh:1:dir:1111111111111111111111111111111111111111"
    runner = CliRunner()
    result = runner.invoke(
        graph_cli_group, ["find-context", "-g", graph_grpc_server, "-c", swhid]
    )
    expected_fqswhid = (
        "Error from the GRPC API call: "
        "Unknown SWHID: swh:1:dir:1111111111111111111111111111111111111111\n"
    )
    assert result.exit_code == 0, result
    assert result.output == expected_fqswhid, result.output


def test_find_context_invalid(graph_grpc_server):
    swhid_1 = ""
    swhid_2 = "asdfasdfasdf"
    runner = CliRunner()
    result_1 = runner.invoke(
        graph_cli_group, ["find-context", "-g", graph_grpc_server, "-c", swhid_1]
    )
    result_2 = runner.invoke(
        graph_cli_group, ["find-context", "-g", graph_grpc_server, "-c", swhid_2]
    )
    assert result_1.exit_code == 0, result_1
    assert result_2.exit_code == 0, result_2
    assert f"'{swhid_1}' is not a valid SWHID" in result_1.output
    assert f"'{swhid_2}' is not a valid SWHID" in result_2.output


def test_find_context_dangling(graph_grpc_server):
    swhid = "swh:1:rel:0000000000000000000000000000000000000019"
    runner = CliRunner()
    result = runner.invoke(
        graph_cli_group, ["find-context", "-g", graph_grpc_server, "-c", swhid]
    )
    expected_fqswhid = (
        "Error from the GRPC API call: "
        "Could not find a path from the sources "
        "(swh:1:rel:0000000000000000000000000000000000000019) to any matching target\n"
    )
    assert result.exit_code == 0, result
    assert result.output == expected_fqswhid, result.output



# Copyright (c) 2022-2025 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import Counter
import hashlib
import re
import time

from google.protobuf.field_mask_pb2 import FieldMask

from swh.graph.example_dataset import DATASET
from swh.graph.grpc.swhgraph_pb2 import (
    GraphDirection,
    NodeFilter,
    StatsRequest,
    TraversalRequest,
)

TEST_ORIGIN_ID = "swh:1:ori:{}".format(
    hashlib.sha1(b"https://example.com/swh/graph").hexdigest()
)


def test_stats_statsd(graph_grpc_stub, graph_statsd_server):
    graph_statsd_server.datagrams = []
    graph_statsd_server.new_datagram.clear()

    graph_grpc_stub.Stats(StatsRequest())

    for _ in range(4):  # usually a single iteration is needed, but may need up to 5
        graph_statsd_server.new_datagram.wait(timeout=1)
        datagrams = [dg.decode() for dg in sorted(graph_statsd_server.datagrams)]
        if len(datagrams) == 4:
            break
        time.sleep(0.01)

    datagrams_by_key = dict(datagram.split(":", 1) for datagram in datagrams)
    assert set(datagrams_by_key) == {
        "swh_graph_grpc_server.frames_total",
        "swh_graph_grpc_server.requests_total",
        "swh_graph_grpc_server.response_wall_time_ms",
        "swh_graph_grpc_server.streaming_wall_time_ms",
    }

    assert datagrams_by_key["swh_graph_grpc_server.frames_total"] == (
        "2|c|#path:/swh.graph.TraversalService/Stats,status:200"
    )

    assert datagrams_by_key["swh_graph_grpc_server.requests_total"] == (
        "1|c|#path:/swh.graph.TraversalService/Stats,status:200"
    )

    assert re.match(
        "[0-9]{1,2}|ms|#path:/swh.graph.TraversalService/Stats,status:200",
        datagrams_by_key["swh_graph_grpc_server.response_wall_time_ms"],
    )

    assert re.match(
        "[0-9]{1,2}|ms|#path:/swh.graph.TraversalService/Stats,status:200",
        datagrams_by_key["swh_graph_grpc_server.streaming_wall_time_ms"],
    )


def test_stats(graph_grpc_stub):
    stats = graph_grpc_stub.Stats(StatsRequest())
    assert stats.num_nodes == 24
    assert stats.num_edges == 28
    assert isinstance(stats.compression_ratio, float)
    assert isinstance(stats.bits_per_node, float)
    assert isinstance(stats.bits_per_edge, float)
    assert isinstance(stats.avg_locality, float)
    assert stats.indegree_min == 0
    assert stats.indegree_max == 4
    assert isinstance(stats.indegree_avg, float)
    assert stats.outdegree_min == 0
    assert stats.outdegree_max == 3
    assert isinstance(stats.outdegree_avg, float)
    assert stats.export_started_at == 1669888200
    assert stats.export_ended_at == 1669899600
    assert stats.num_nodes_by_type == dict(
        Counter(
            node.swhid().object_type.value for node in DATASET if hasattr(node, "swhid")
        )
    )
    # TODO: auto-generate this
    assert stats.num_arcs_by_type == {
        "ori:snp": 2,
        "snp:rel": 3,
        "snp:rev": 2,
        "rel:rev": 3,
        "rev:rev": 3,
        "rev:dir": 4,
        "dir:dir": 3,
        "dir:cnt": 8,
    }


def test_leaves(graph_grpc_stub):
    request = graph_grpc_stub.Traverse(
        TraversalRequest(
            src=[TEST_ORIGIN_ID],
            mask=FieldMask(paths=["swhid"]),
            return_nodes=NodeFilter(types="cnt"),
        )
    )
    actual = [node.swhid for node in request]
    expected = [
        "swh:1:cnt:0000000000000000000000000000000000000001",
        "swh:1:cnt:0000000000000000000000000000000000000004",
        "swh:1:cnt:0000000000000000000000000000000000000005",
        "swh:1:cnt:0000000000000000000000000000000000000007",
    ]
    assert set(actual) == set(expected)


def test_neighbors(graph_grpc_stub):
    request = graph_grpc_stub.Traverse(
        TraversalRequest(
            src=["swh:1:rev:0000000000000000000000000000000000000009"],
            direction=GraphDirection.BACKWARD,
            mask=FieldMask(paths=["swhid"]),
            min_depth=1,
            max_depth=1,
        )
    )
    actual = [node.swhid for node in request]
    expected = [
        "swh:1:snp:0000000000000000000000000000000000000022",
        "swh:1:snp:0000000000000000000000000000000000000020",
        "swh:1:rel:0000000000000000000000000000000000000010",
        "swh:1:rev:0000000000000000000000000000000000000013",
    ]
    assert set(actual) == set(expected)


def test_visit_nodes(graph_grpc_stub):
    request = graph_grpc_stub.Traverse(
        TraversalRequest(
            src=["swh:1:rel:0000000000000000000000000000000000000010"],
            mask=FieldMask(paths=["swhid"]),
            edges="rel:rev,rev:rev",
        )
    )
    actual = [node.swhid for node in request]
    expected = [
        "swh:1:rel:0000000000000000000000000000000000000010",
        "swh:1:rev:0000000000000000000000000000000000000009",
        "swh:1:rev:0000000000000000000000000000000000000003",
    ]
    assert set(actual) == set(expected)


def test_visit_nodes_filtered(graph_grpc_stub):
    request = graph_grpc_stub.Traverse(
        TraversalRequest(
            src=["swh:1:rel:0000000000000000000000000000000000000010"],
            mask=FieldMask(paths=["swhid"]),
            return_nodes=NodeFilter(types="dir"),
        )
    )
    actual = [node.swhid for node in request]
    expected = [
        "swh:1:dir:0000000000000000000000000000000000000002",
        "swh:1:dir:0000000000000000000000000000000000000008",
        "swh:1:dir:0000000000000000000000000000000000000006",
    ]
    assert set(actual) == set(expected)


def test_visit_nodes_filtered_star(graph_grpc_stub):
    request = graph_grpc_stub.Traverse(
        TraversalRequest(
            src=["swh:1:rel:0000000000000000000000000000000000000010"],
            mask=FieldMask(paths=["swhid"]),
        )
    )
    actual = [node.swhid for node in request]
    expected = [
        "swh:1:rel:0000000000000000000000000000000000000010",
        "swh:1:rev:0000000000000000000000000000000000000009",
        "swh:1:rev:0000000000000000000000000000000000000003",
        "swh:1:dir:0000000000000000000000000000000000000002",
        "swh:1:cnt:0000000000000000000000000000000000000001",
        "swh:1:dir:0000000000000000000000000000000000000008",
        "swh:1:cnt:0000000000000000000000000000000000000007",
        "swh:1:dir:0000000000000000000000000000000000000006",
        "swh:1:cnt:0000000000000000000000000000000000000004",
        "swh:1:cnt:0000000000000000000000000000000000000005",
    ]
    assert set(actual) == set(expected)



# Copyright (C) 2022-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from hashlib import sha1

from google.protobuf.field_mask_pb2 import FieldMask
import grpc

from swh.graph.grpc.swhgraph_pb2 import (
    FindPathBetweenRequest,
    FindPathToRequest,
    NodeFilter,
    TraversalRequest,
)
from swh.graph.grpc.swhgraph_pb2_grpc import TraversalServiceStub
from swh.model.cli import swhid_of_file
from swh.model.exceptions import ValidationError

# documentation: https://docs.softwareheritage.org/devel/apidoc/swh.model.swhids.html
from swh.model.swhids import CoreSWHID, ExtendedObjectType, ExtendedSWHID

# global variable holding headers parameters
headers: dict = {}
swhcli: dict = {}
verbose = False

GRAPH_GRPC_SERVER = "localhost:50091"


def fqswhid_of_traversal(response):
    # Build the Fully qualified SWHID
    global verbose
    fqswhid = []
    coreswhidtype = ""
    needrevision = True
    needrelease = True
    path = []
    url = ""
    for node in response.node:
        # response contains the nodes in the order content -> dir -> revision
        # -> release -> snapshot -> origin
        if verbose:
            print(" examining node: " + node.swhid)
        if url == "":
            url = node.ori.url
        parsedid = ExtendedSWHID.from_string(node.swhid)
        if parsedid.object_type == ExtendedObjectType.CONTENT:
            # print(parsedid.object_type)
            # print(swhcli.get(node.swhid))
            pathids = [
                label.name.decode()
                for successor in node.successor
                for label in successor.label
            ]
            path.insert(0, pathids[0])  # raises exception if pathids is empty!
            fqswhid.append(node.swhid)
            coreswhidtype = "cnt"
        if parsedid.object_type == ExtendedObjectType.DIRECTORY:
            # print(parsedid.object_type)
            # print(swhcli.get(node.swhid))
            if fqswhid:  # empty list signals coreswhid not found yet
                pathids = [
                    label.name.decode()
                    for successor in node.successor
                    for label in successor.label
                ]
                path.insert(0, pathids[0])  # raises exception if pathids is empty!
            else:
                fqswhid.append(node.swhid)
                coreswhidtype = "dir"
        if parsedid.object_type == ExtendedObjectType.REVISION:
            if fqswhid:  # empty list signals coreswhid not found yet
                if needrevision:
                    revision = node.swhid
                    needrevision = False
            else:
                fqswhid.append(node.swhid)
                coreswhidtype = "rev"
        if parsedid.object_type == ExtendedObjectType.RELEASE:
            if fqswhid:  # empty list signals coreswhid not found yet
                if needrelease:
                    release = node.swhid
                    needrelease = False
            else:
                fqswhid.append(node.swhid)
                coreswhidtype = "rel"
        if parsedid.object_type == ExtendedObjectType.SNAPSHOT:
            if fqswhid:  # empty list signals coreswhid not found yet
                snapshot = node.swhid
            else:
                fqswhid.append(node.swhid)
                coreswhidtype = "snp"
    # Now we have all the elements to print a FQ SWHID
    # We could also build and return a swh.model.swhids.QualifiedSWHID
    if coreswhidtype == "cnt" or coreswhidtype == "dir":
        fqswhid.append("path=/" + "/".join(path))
        if not needrevision:
            fqswhid.append("anchor=" + revision)
        elif not needrelease:
            fqswhid.append("anchor=" + release)
        if snapshot:
            fqswhid.append("visit=" + snapshot)
    elif coreswhidtype == "rev" or coreswhidtype == "rel":
        if snapshot:
            fqswhid.append("visit=" + snapshot)
    if url:
        fqswhid.append("origin=" + url)
    return ";".join(fqswhid)


def main(
    content_swhid,
    origin_url,
    all_origins,
    random_origin,
    filename,
    graph_grpc_server,
    fqswhid,
    trace,
):
    global headers
    global swhcli
    global verbose
    verbose = trace

    # Check if content SWHID is valid
    try:
        CoreSWHID.from_string(content_swhid)
    except ValidationError:
        print(f"Error: '{content_swhid}' is not a valid SWHID")
        return

    with grpc.insecure_channel(graph_grpc_server) as channel:
        client = TraversalServiceStub(channel)
        field_mask = FieldMask(
            paths=[
                "node.swhid",
                "node.ori.url",
                "node.successor.swhid",
                "node.successor.label.name",
            ]
        )

        try:
            if filename:
                content_swhid = str(swhid_of_file(filename))

            # Traversal request: get all origins
            if all_origins:
                random_origin = False
                response = client.Traverse(
                    TraversalRequest(
                        src=[content_swhid],
                        edges="cnt:dir,dir:dir,dir:rev,rev:rev,rev:snp,rev:rel,rel:snp,snp:ori",
                        direction="BACKWARD",
                        return_nodes=NodeFilter(types="ori"),
                        mask=field_mask,
                    )
                )
                for node in response:
                    if fqswhid:
                        response = client.FindPathBetween(
                            FindPathBetweenRequest(
                                src=[content_swhid],
                                dst=[node.swhid],
                                direction="BACKWARD",
                                mask=field_mask,
                            )
                        )
                        print(fqswhid_of_traversal(response))
                    else:
                        print(node.ori.url)

            # Traversal request to a (random) origin
            if random_origin:
                response = client.FindPathTo(
                    FindPathToRequest(
                        src=[content_swhid],
                        target=NodeFilter(types="ori"),
                        direction="BACKWARD",
                        mask=field_mask,
                    )
                )
                if fqswhid:
                    print(fqswhid_of_traversal(response))
                else:
                    for node in response.node:
                        print(node.ori.url)

            # Traversal request to a given origin URL
            if origin_url:
                response = client.FindPathBetween(
                    FindPathBetweenRequest(
                        src=[content_swhid],
                        dst=[
                            str(
                                ExtendedSWHID(
                                    object_type=ExtendedObjectType.ORIGIN,
                                    object_id=bytes.fromhex(
                                        sha1(bytes(origin_url, "UTF-8")).hexdigest()
                                    ),
                                )
                            )
                        ],
                        direction="BACKWARD",
                        mask=field_mask,
                    )
                )
                print(fqswhid_of_traversal(response))

        except grpc.RpcError as e:
            print("Error from the GRPC API call: {}".format(e.details()))
            if filename:
                print(filename + " has SWHID " + content_swhid)
        except Exception as e:
            print("Unexpected error occurred: {}".format(e))



# Copyright (C) 2021-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import functools
import inspect
import itertools
import re
import statistics
from typing import (
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

from swh.model.swhids import CoreSWHID, ExtendedSWHID, ValidationError

from .http_client import GraphArgumentException

_NODE_TYPES = "ori|snp|rel|rev|dir|cnt"
NODES_RE = re.compile(rf"(\*|{_NODE_TYPES})")
EDGES_RE = re.compile(rf"(\*|{_NODE_TYPES}):(\*|{_NODE_TYPES})")


T = TypeVar("T", bound=Callable)
SWHIDlike = Union[CoreSWHID, ExtendedSWHID, str]


def check_arguments(f: T) -> T:
    """Decorator for generic argument checking for methods of NaiveClient.
    Checks ``src`` is a valid and known SWHID, and ``edges`` has the right format."""
    signature = inspect.signature(f)

    @functools.wraps(f)
    def newf(*args, **kwargs):
        __tracebackhide__ = True  # for pytest
        try:
            bound_args = signature.bind(*args, **kwargs)
        except TypeError as e:
            # rethrow the exception from here so pytest doesn't flood the terminal
            # with signature.bind's call stack.
            raise TypeError(*e.args) from None
        self = bound_args.arguments["self"]

        src = bound_args.arguments.get("src")
        if src:
            self._check_swhid(src)

        edges = bound_args.arguments.get("edges")
        if edges:
            if edges != "*" and not EDGES_RE.match(edges):
                raise GraphArgumentException(f"invalid edge restriction: {edges}")

        return_types = bound_args.arguments.get("return_types")
        if return_types:
            if not NODES_RE.match(return_types):
                raise GraphArgumentException(
                    f"invalid return_types restriction: {return_types}"
                )

        return f(*args, **kwargs)

    return newf  # type: ignore


def filter_node_types(node_types: str, nodes: Iterable[str]) -> Iterator[str]:
    if node_types == "*":
        yield from nodes
    else:
        prefixes = tuple(f"swh:1:{type_}:" for type_ in node_types.split(","))
        for node in nodes:
            if node.startswith(prefixes):
                yield node


class NaiveClient:
    """An alternative implementation of the graph server, written in
    pure-python and meant for simulating it in other components' test cases;
    constructed from a list of nodes and (directed) edges, both represented as
    SWHIDs.

    It is NOT meant to be efficient in any way; only to be a very simple
    implementation that provides the same behavior.

    >>> nodes = [
    ...     "swh:1:rev:1111111111111111111111111111111111111111",
    ...     "swh:1:rev:2222222222222222222222222222222222222222",
    ...     "swh:1:rev:3333333333333333333333333333333333333333",
    ... ]
    >>> edges = [
    ...     (
    ...         "swh:1:rev:1111111111111111111111111111111111111111",
    ...         "swh:1:rev:2222222222222222222222222222222222222222",
    ...     ),
    ...     (
    ...         "swh:1:rev:2222222222222222222222222222222222222222",
    ...         "swh:1:rev:3333333333333333333333333333333333333333",
    ...     ),
    ... ]
    >>> c = NaiveClient(nodes=nodes, edges=edges)
    >>> list(c.leaves("swh:1:rev:1111111111111111111111111111111111111111"))
    ['swh:1:rev:3333333333333333333333333333333333333333']
    """

    def __init__(
        self, *, nodes: List[SWHIDlike], edges: List[Tuple[SWHIDlike, SWHIDlike]]
    ):
        self.graph = Graph(nodes, edges)

    def _check_swhid(self, swhid):
        try:
            ExtendedSWHID.from_string(swhid)
        except ValidationError as e:
            raise GraphArgumentException(*e.args) from None
        if swhid not in self.graph.nodes:
            raise GraphArgumentException(f"SWHID not found: {swhid}")

    def stats(self) -> Dict:
        return {
            "num_nodes": len(self.graph.nodes),
            "num_edges": sum(map(len, self.graph.forward_edges.values())),
            "compression_ratio": 1.0,
            "bits_per_edge": 100.0,
            "bits_per_node": 100.0,
            "avg_locality": 0.0,
            "indegree_min": min(map(len, self.graph.backward_edges.values())),
            "indegree_max": max(map(len, self.graph.backward_edges.values())),
            "indegree_avg": statistics.mean(
                map(len, self.graph.backward_edges.values())
            ),
            "outdegree_min": min(map(len, self.graph.forward_edges.values())),
            "outdegree_max": max(map(len, self.graph.forward_edges.values())),
            "outdegree_avg": statistics.mean(
                map(len, self.graph.forward_edges.values())
            ),
            "export_started_at": 1669888200,
            "export_ended_at": 1669899600,
        }

    @check_arguments
    def leaves(
        self,
        src: str,
        edges: str = "*",
        direction: str = "forward",
        max_edges: int = 0,
        return_types: str = "*",
        max_matching_nodes: int = 0,
    ) -> Iterator[str]:
        if max_edges > 0:
            raise NotImplementedError("max_edges")  # TODO
        leaves = filter_node_types(
            return_types,
            [
                node
                for node in self.graph.get_subgraph(src, edges, direction)
                if not self.graph.get_filtered_neighbors(node, edges, direction)
            ],
        )

        if max_matching_nodes > 0:
            leaves = itertools.islice(leaves, max_matching_nodes)

        return leaves

    @check_arguments
    def neighbors(
        self,
        src: str,
        edges: str = "*",
        direction: str = "forward",
        max_edges: int = 0,
        return_types: str = "*",
        max_matching_nodes: int = 0,
    ) -> Iterator[str]:
        yield from filter_node_types(
            return_types,
            self.graph.get_filtered_neighbors(
                src, edges, direction, max_edges, max_matching_nodes
            ),
        )

    @check_arguments
    def visit_nodes(
        self,
        src: str,
        edges: str = "*",
        direction: str = "forward",
        max_edges: int = 0,
        return_types: str = "*",
        max_matching_nodes: int = 0,
    ) -> Iterator[str]:
        if max_edges > 0:
            raise NotImplementedError("max_edges")  # TODO
        res = filter_node_types(
            return_types, self.graph.get_subgraph(src, edges, direction)
        )
        if max_matching_nodes > 0:
            res = itertools.islice(res, max_matching_nodes)
        return res

    @check_arguments
    def visit_edges(
        self, src: str, edges: str = "*", direction: str = "forward", max_edges: int = 0
    ) -> Iterator[Tuple[str, str]]:
        if max_edges == 0:
            max_edges = None  # type: ignore
        else:
            max_edges -= 1
        yield from list(self.graph.iter_edges_dfs(direction, edges, src))[:max_edges]

    @check_arguments
    def visit_paths(
        self, src: str, edges: str = "*", direction: str = "forward", max_edges: int = 0
    ) -> Iterator[List[str]]:
        if max_edges > 0:
            raise NotImplementedError("max_edges")  # TODO
        for path in self.graph.iter_paths_dfs(direction, edges, src):
            if path[-1] in self.leaves(src, edges, direction):
                yield list(path)

    @check_arguments
    def walk(
        self,
        src: str,
        dst: str,
        edges: str = "*",
        traversal: str = "dfs",
        direction: str = "forward",
        limit: Optional[int] = None,
    ) -> Iterator[str]:
        # TODO: implement algo="bfs"
        # TODO: limit
        match_path: Callable[[str], bool]
        if ":" in dst:
            match_path = dst.__eq__
            self._check_swhid(dst)
        else:
            match_path = lambda node: node.startswith(f"swh:1:{dst}:")  # noqa
        for path in self.graph.iter_paths_dfs(direction, edges, src):
            if match_path(path[-1]):
                if not limit:
                    # 0 or None
                    yield from path
                elif limit > 0:
                    yield from path[0:limit]
                else:
                    yield from path[limit:]

    @check_arguments
    def random_walk(
        self,
        src: str,
        dst: str,
        edges: str = "*",
        direction: str = "forward",
        limit: Optional[int] = None,
    ):
        # TODO: limit
        yield from self.walk(src, dst, edges, "dfs", direction, limit)

    @check_arguments
    def count_leaves(
        self,
        src: str,
        edges: str = "*",
        direction: str = "forward",
        max_matching_nodes: int = 0,
    ) -> int:
        return len(
            list(
                self.leaves(
                    src, edges, direction, max_matching_nodes=max_matching_nodes
                )
            )
        )

    @check_arguments
    def count_neighbors(
        self, src: str, edges: str = "*", direction: str = "forward"
    ) -> int:
        return len(self.graph.get_filtered_neighbors(src, edges, direction))

    @check_arguments
    def count_visit_nodes(
        self,
        src: str,
        edges: str = "*",
        direction: str = "forward",
        max_matching_nodes: int = 0,
    ) -> int:
        res = len(self.graph.get_subgraph(src, edges, direction))
        if max_matching_nodes > 0:
            res = min(max_matching_nodes, res)
        return res


class Graph:
    def __init__(
        self, nodes: List[SWHIDlike], edges: List[Tuple[SWHIDlike, SWHIDlike]]
    ):
        self.nodes = [str(node) for node in nodes]
        self.forward_edges: Dict[str, List[str]] = {}
        self.backward_edges: Dict[str, List[str]] = {}
        for node in nodes:
            self.forward_edges[str(node)] = []
            self.backward_edges[str(node)] = []
        for src, dst in edges:
            self.forward_edges[str(src)].append(str(dst))
            self.backward_edges[str(dst)].append(str(src))

    def get_filtered_neighbors(
        self,
        src: str,
        edges_fmt: str,
        direction: str,
        max_edges: int = 0,
        max_matching_nodes: int = 0,
    ) -> Set[str]:
        if direction == "forward":
            edges = self.forward_edges
        elif direction == "backward":
            edges = self.backward_edges
        else:
            raise GraphArgumentException(f"invalid direction: {direction}")

        neighbors = edges.get(src, [])
        filtered_neighbors: Set[str] = set()

        if edges_fmt == "*":
            if max_edges:
                filtered_neighbors = set(neighbors[0:max_edges])
            else:
                filtered_neighbors = set(neighbors)
        else:
            accessed_edges = 0
            for edges_fmt_item in edges_fmt.split(","):
                if max_edges > 0:
                    accessed_edges += 1
                    if accessed_edges > max_edges:
                        break
                (src_fmt, dst_fmt) = edges_fmt_item.split(":")
                if src_fmt != "*" and not src.startswith(f"swh:1:{src_fmt}:"):
                    continue
                if dst_fmt == "*":
                    filtered_neighbors.update(neighbors)
                else:
                    prefix = f"swh:1:{dst_fmt}:"
                    filtered_neighbors.update(
                        n for n in neighbors if n.startswith(prefix)
                    )

        # Apply max_matching_nodes to the set of filtered neighbors
        if not max_matching_nodes or len(filtered_neighbors) <= max_matching_nodes:
            return filtered_neighbors

        return set(list(filtered_neighbors)[:max_matching_nodes])

    def get_subgraph(self, src: str, edges_fmt: str, direction: str) -> Set[str]:
        seen = set()
        to_visit = {src}
        while to_visit:
            node = to_visit.pop()
            seen.add(node)
            neighbors = set(self.get_filtered_neighbors(node, edges_fmt, direction))
            new_nodes = neighbors - seen
            to_visit.update(new_nodes)

        return seen

    def iter_paths_dfs(
        self, direction: str, edges_fmt: str, src: str
    ) -> Iterator[Tuple[str, ...]]:
        for path, node in DfsSubgraphIterator(self, direction, edges_fmt, src):
            yield path + (node,)

    def iter_edges_dfs(
        self, direction: str, edges_fmt: str, src: str
    ) -> Iterator[Tuple[str, str]]:
        for path, node in DfsSubgraphIterator(self, direction, edges_fmt, src):
            if len(path) > 0:
                yield (path[-1], node)


class SubgraphIterator(Iterator[Tuple[Tuple[str, ...], str]]):
    def __init__(self, graph: Graph, direction: str, edges_fmt: str, src: str):
        self.graph = graph
        self.direction = direction
        self.edges_fmt = edges_fmt
        self.seen: Set[str] = set()
        self.src = src

    def more_work(self) -> bool:
        raise NotImplementedError()

    def pop(self) -> Tuple[Tuple[str, ...], str]:
        raise NotImplementedError()

    def push(self, new_path: Tuple[str, ...], neighbor: str) -> None:
        raise NotImplementedError()

    def __next__(self) -> Tuple[Tuple[str, ...], str]:
        # Stores (path, next_node)
        if not self.more_work():
            raise StopIteration()

        (path, node) = self.pop()

        new_path = path + (node,)

        if node not in self.seen:
            neighbors = self.graph.get_filtered_neighbors(
                node, self.edges_fmt, self.direction
            )

            # We want to visit the first neighbor first, and to_visit is a stack;
            # so we need to reversed() the list of neighbors to get it on top
            # of the stack.
            for neighbor in reversed(list(neighbors)):
                self.push(new_path, neighbor)

        self.seen.add(node)
        return (path, node)


class DfsSubgraphIterator(SubgraphIterator):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.to_visit: List[Tuple[Tuple[str, ...], str]] = [((), self.src)]

    def more_work(self) -> bool:
        return bool(self.to_visit)

    def pop(self) -> Tuple[Tuple[str, ...], str]:
        return self.to_visit.pop()

    def push(self, new_path: Tuple[str, ...], neighbor: str) -> None:
        self.to_visit.append((new_path, neighbor))



# Copyright (c) 2022-2025 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import hashlib

from google.protobuf.field_mask_pb2 import FieldMask

from swh.graph.grpc.swhgraph_pb2 import (
    EdgeLabel,
    FindPathBetweenRequest,
    FindPathToRequest,
    GraphDirection,
    Node,
    NodeFilter,
    Successor,
    TraversalRequest,
)

TEST_ORIGIN_ID = "swh:1:ori:{}".format(
    hashlib.sha1(b"https://example.com/swh/graph").hexdigest()
)
TEST_ORIGIN_ID2 = "swh:1:ori:{}".format(
    hashlib.sha1(b"https://example.com/swh/graph2").hexdigest()
)


def test_traverse_forward_labels(graph_grpc_stub):
    request = graph_grpc_stub.Traverse(
        TraversalRequest(
            src=[TEST_ORIGIN_ID],
            mask=FieldMask(paths=["swhid", "successor.swhid", "successor.label"]),
            direction=GraphDirection.FORWARD,
        )
    )
    expected = [
        Node(
            swhid="swh:1:cnt:0000000000000000000000000000000000000001",
            successor=None,
            cnt=None,
        ),
        Node(
            swhid="swh:1:cnt:0000000000000000000000000000000000000004",
            successor=None,
            cnt=None,
        ),
        Node(
            swhid="swh:1:cnt:0000000000000000000000000000000000000005",
            successor=None,
            cnt=None,
        ),
        Node(
            swhid="swh:1:cnt:0000000000000000000000000000000000000007",
            successor=None,
            cnt=None,
        ),
        Node(
            swhid="swh:1:dir:0000000000000000000000000000000000000002",
            successor=[
                Successor(
                    swhid="swh:1:cnt:0000000000000000000000000000000000000001",
                    label=[EdgeLabel(name=b"README.md", permission=0o100644)],
                )
            ],
        ),
        Node(
            swhid="swh:1:dir:0000000000000000000000000000000000000006",
            successor=[
                Successor(
                    swhid="swh:1:cnt:0000000000000000000000000000000000000004",
                    label=[EdgeLabel(name=b"README.rst", permission=0o100644)],
                ),
                Successor(
                    swhid="swh:1:cnt:0000000000000000000000000000000000000005",
                    label=[EdgeLabel(name=b"main.py", permission=0o100644)],
                ),
            ],
        ),
        Node(
            swhid="swh:1:dir:0000000000000000000000000000000000000008",
            successor=[
                Successor(
                    swhid="swh:1:cnt:0000000000000000000000000000000000000001",
                    label=[EdgeLabel(name=b"README.md", permission=0o100644)],
                ),
                Successor(
                    swhid="swh:1:cnt:0000000000000000000000000000000000000007",
                    label=[EdgeLabel(name=b"parser.c", permission=0o100644)],
                ),
                Successor(
                    swhid="swh:1:dir:0000000000000000000000000000000000000006",
                    label=[EdgeLabel(name=b"tests", permission=0o100755)],
                ),
            ],
        ),
        Node(
            swhid=TEST_ORIGIN_ID,
            successor=[
                Successor(
                    swhid="swh:1:snp:0000000000000000000000000000000000000020",
                    label=[EdgeLabel(visit_timestamp=1367900441, is_full_visit=True)],
                ),
            ],
            ori=None,
        ),
        Node(
            swhid="swh:1:rel:0000000000000000000000000000000000000010",
            successor=[
                Successor(
                    swhid="swh:1:rev:0000000000000000000000000000000000000009",
                    label=[],
                ),
            ],
            rel=None,
        ),
        Node(
            swhid="swh:1:rev:0000000000000000000000000000000000000003",
            successor=[
                Successor(
                    swhid="swh:1:dir:0000000000000000000000000000000000000002",
                    label=[],
                ),
            ],
            rev=None,
        ),
        Node(
            swhid="swh:1:rev:0000000000000000000000000000000000000009",
            successor=[
                Successor(
                    swhid="swh:1:dir:0000000000000000000000000000000000000008",
                    label=[],
                ),
                Successor(
                    swhid="swh:1:rev:0000000000000000000000000000000000000003",
                    label=[],
                ),
            ],
            rev=None,
        ),
        Node(
            swhid="swh:1:snp:0000000000000000000000000000000000000020",
            successor=[
                Successor(
                    swhid="swh:1:rel:0000000000000000000000000000000000000010",
                    label=[
                        EdgeLabel(name=b"refs/tags/v1.0"),
                    ],
                ),
                Successor(
                    swhid="swh:1:rev:0000000000000000000000000000000000000009",
                    label=[
                        EdgeLabel(name=b"refs/heads/master"),
                    ],
                ),
            ],
        ),
    ]
    actual = list(request)
    actual.sort(key=lambda node: node.swhid)
    for node in actual:
        node.successor.sort(key=lambda successor: successor.swhid)
    assert actual == expected


def test_traverse_backward_labels(graph_grpc_stub):
    request = graph_grpc_stub.Traverse(
        TraversalRequest(
            src=["swh:1:cnt:0000000000000000000000000000000000000015"],
            mask=FieldMask(paths=["swhid", "successor.swhid", "successor.label"]),
            direction=GraphDirection.BACKWARD,
        )
    )
    expected = [
        Node(
            swhid="swh:1:cnt:0000000000000000000000000000000000000015",
            successor=[
                Successor(
                    swhid="swh:1:dir:0000000000000000000000000000000000000016",
                    label=[EdgeLabel(name=b"TODO.txt", permission=0o100644)],
                )
            ],
            cnt=None,
        ),
        Node(
            swhid="swh:1:dir:0000000000000000000000000000000000000016",
            successor=[
                Successor(
                    swhid="swh:1:dir:0000000000000000000000000000000000000017",
                    label=[EdgeLabel(name=b"old", permission=0o100755)],
                ),
            ],
        ),
        Node(
            swhid="swh:1:dir:0000000000000000000000000000000000000017",
            successor=[
                Successor(
                    swhid="swh:1:rev:0000000000000000000000000000000000000018",
                    label=[],
                ),
            ],
        ),
        Node(
            swhid=TEST_ORIGIN_ID2,
            successor=None,
            ori=None,
        ),
        Node(
            swhid="swh:1:rel:0000000000000000000000000000000000000019",
            successor=[],
            rel=None,
        ),
        Node(
            swhid="swh:1:rel:0000000000000000000000000000000000000021",
            successor=[
                Successor(
                    swhid="swh:1:snp:0000000000000000000000000000000000000022",
                    label=[
                        EdgeLabel(name=b"refs/tags/v2.0-anonymous"),
                    ],
                ),
            ],
            rel=None,
        ),
        Node(
            swhid="swh:1:rev:0000000000000000000000000000000000000018",
            successor=[
                Successor(
                    swhid="swh:1:rel:0000000000000000000000000000000000000019",
                    label=[],
                ),
                Successor(
                    swhid="swh:1:rel:0000000000000000000000000000000000000021",
                    label=[],
                ),
            ],
            rev=None,
        ),
        Node(
            swhid="swh:1:snp:0000000000000000000000000000000000000022",
            successor=[
                Successor(
                    swhid=TEST_ORIGIN_ID2,
                    label=[EdgeLabel(visit_timestamp=1367900441, is_full_visit=True)],
                ),
            ],
        ),
    ]
    actual = list(request)
    actual.sort(key=lambda node: node.swhid)
    for node in actual:
        node.successor.sort(key=lambda successor: successor.swhid)
    assert actual == expected


def test_findpathto_forward_labels(graph_grpc_stub):
    request = graph_grpc_stub.FindPathTo(
        FindPathToRequest(
            src=[TEST_ORIGIN_ID],
            target=NodeFilter(types="rev"),
            mask=FieldMask(
                paths=["swhid", "node.successor.swhid", "node.successor.label"]
            ),
            direction=GraphDirection.FORWARD,
        )
    )
    for node in request.node:
        node.successor.sort(key=lambda successor: successor.swhid)
    assert_ori_to_rev(request.node)


def test_findpathbetween_forward_labels(graph_grpc_stub):
    request = graph_grpc_stub.FindPathBetween(
        FindPathBetweenRequest(
            src=[TEST_ORIGIN_ID],
            dst=["swh:1:rev:0000000000000000000000000000000000000009"],
            mask=FieldMask(
                paths=["swhid", "node.successor.swhid", "node.successor.label"]
            ),
            direction=GraphDirection.FORWARD,
        )
    )
    for node in request.node:
        node.successor.sort(key=lambda successor: successor.swhid)
    assert_ori_to_rev(request.node)


def assert_ori_to_rev(path):
    assert path == [
        Node(
            swhid=TEST_ORIGIN_ID,
            successor=[
                Successor(
                    swhid="swh:1:snp:0000000000000000000000000000000000000020",
                    label=[EdgeLabel(visit_timestamp=1367900441, is_full_visit=True)],
                ),
            ],
            ori=None,
        ),
        Node(
            swhid="swh:1:snp:0000000000000000000000000000000000000020",
            successor=[
                Successor(
                    swhid="swh:1:rel:0000000000000000000000000000000000000010",
                    label=[EdgeLabel(name=b"refs/tags/v1.0")],
                ),
                Successor(
                    swhid="swh:1:rev:0000000000000000000000000000000000000009",
                    label=[EdgeLabel(name=b"refs/heads/master")],
                ),
            ],
        ),
        Node(
            swhid="swh:1:rev:0000000000000000000000000000000000000009",
            successor=[
                Successor(
                    swhid="swh:1:dir:0000000000000000000000000000000000000008",
                ),
                Successor(
                    swhid="swh:1:rev:0000000000000000000000000000000000000003",
                ),
            ],
            rev=None,
        ),
    ]


def test_findpathto_backward_labels(graph_grpc_stub):
    request = graph_grpc_stub.FindPathTo(
        FindPathToRequest(
            src=["swh:1:rel:0000000000000000000000000000000000000021"],
            target=NodeFilter(types="ori"),
            mask=FieldMask(
                paths=["swhid", "node.successor.swhid", "node.successor.label"]
            ),
            direction=GraphDirection.BACKWARD,
        )
    )
    for node in request.node:
        node.successor.sort(key=lambda successor: successor.swhid)
    assert_rel_to_ori(request.node)


def test_findpathbetween_backward_labels(graph_grpc_stub):
    request = graph_grpc_stub.FindPathBetween(
        FindPathBetweenRequest(
            src=["swh:1:rel:0000000000000000000000000000000000000021"],
            dst=[TEST_ORIGIN_ID2],
            mask=FieldMask(
                paths=["swhid", "node.successor.swhid", "node.successor.label"]
            ),
            direction=GraphDirection.BACKWARD,
        )
    )
    for node in request.node:
        node.successor.sort(key=lambda successor: successor.swhid)
    assert_rel_to_ori(request.node)


def assert_rel_to_ori(path):
    assert path == [
        Node(
            swhid="swh:1:rel:0000000000000000000000000000000000000021",
            successor=[
                Successor(
                    swhid="swh:1:snp:0000000000000000000000000000000000000022",
                    label=[EdgeLabel(name=b"refs/tags/v2.0-anonymous")],
                ),
            ],
            rel=None,
        ),
        Node(
            swhid="swh:1:snp:0000000000000000000000000000000000000022",
            successor=[
                Successor(
                    swhid=TEST_ORIGIN_ID2,
                    label=[EdgeLabel(visit_timestamp=1367900441, is_full_visit=True)],
                ),
            ],
        ),
        Node(swhid=TEST_ORIGIN_ID2, successor=[], ori=None),
    ]


def test_findpathbetween_common_parent_labels(graph_grpc_stub):
    request = graph_grpc_stub.FindPathBetween(
        FindPathBetweenRequest(
            src=["swh:1:rel:0000000000000000000000000000000000000010"],
            dst=["swh:1:rel:0000000000000000000000000000000000000021"],
            mask=FieldMask(
                paths=["swhid", "node.successor.swhid", "node.successor.label"]
            ),
            direction=GraphDirection.BACKWARD,
            direction_reverse=GraphDirection.BACKWARD,
        )
    )
    for node in request.node:
        node.successor.sort(key=lambda successor: successor.swhid)
    assert request.node == [
        Node(
            swhid="swh:1:rel:0000000000000000000000000000000000000010",
            successor=[
                Successor(
                    swhid="swh:1:snp:0000000000000000000000000000000000000020",
                    label=[EdgeLabel(name=b"refs/tags/v1.0")],
                ),
                Successor(
                    swhid="swh:1:snp:0000000000000000000000000000000000000022",
                    label=[EdgeLabel(name=b"refs/tags/v1.0")],
                ),
            ],
            rel=None,
        ),
        Node(
            swhid="swh:1:snp:0000000000000000000000000000000000000022",
            successor=[
                Successor(
                    swhid=TEST_ORIGIN_ID2,
                    label=[EdgeLabel(visit_timestamp=1367900441, is_full_visit=True)],
                ),
            ],
        ),
        Node(
            swhid="swh:1:rel:0000000000000000000000000000000000000021",
            successor=[
                Successor(
                    swhid="swh:1:snp:0000000000000000000000000000000000000022",
                    label=[EdgeLabel(name=b"refs/tags/v2.0-anonymous")],
                ),
            ],
            rel=None,
        ),
    ]


def test_findpathbetween_common_child_labels(graph_grpc_stub):
    request = graph_grpc_stub.FindPathBetween(
        FindPathBetweenRequest(
            src=["swh:1:dir:0000000000000000000000000000000000000002"],
            dst=["swh:1:dir:0000000000000000000000000000000000000008"],
            mask=FieldMask(
                paths=["swhid", "node.successor.swhid", "node.successor.label"]
            ),
            direction=GraphDirection.FORWARD,
            direction_reverse=GraphDirection.FORWARD,
        )
    )
    for node in request.node:
        node.successor.sort(key=lambda successor: successor.swhid)
    assert request.node == [
        Node(
            swhid="swh:1:dir:0000000000000000000000000000000000000002",
            successor=[
                Successor(
                    swhid="swh:1:cnt:0000000000000000000000000000000000000001",
                    label=[EdgeLabel(name=b"README.md", permission=0o100644)],
                ),
            ],
        ),
        Node(
            swhid="swh:1:cnt:0000000000000000000000000000000000000001",
            successor=[],
            cnt=None,
        ),
        Node(
            swhid="swh:1:dir:0000000000000000000000000000000000000008",
            successor=[
                Successor(
                    swhid="swh:1:cnt:0000000000000000000000000000000000000001",
                    label=[EdgeLabel(name=b"README.md", permission=0o100644)],
                ),
                Successor(
                    swhid="swh:1:cnt:0000000000000000000000000000000000000007",
                    label=[EdgeLabel(name=b"parser.c", permission=0o100644)],
                ),
                Successor(
                    swhid="swh:1:dir:0000000000000000000000000000000000000006",
                    label=[EdgeLabel(name=b"tests", permission=0o100755)],
                ),
            ],
        ),
    ]
